////////////////////////////////////////////////////////////////////////////
// File:	MultiCtrl.cpp
// Version:	2
// Created:	05-Mar-2002
//
// Author:	Paul S. Vickery
// E-mail:	paul@vickeryhome.freeserve.co.uk
//
// Control switchable at run-time to become any one of a number of controls
//
// You are free to use or modify this code, with no restrictions, other than
// you continue to acknowledge me as the original author in this source code,
// or any code derived from it.
//
// If you use this code, or use it as a base for your own code, it would be 
// nice to hear from you simply so I know it's not been a waste of time!
//
// Copyright (c) 2001-2002 Paul S. Vickery
//
////////////////////////////////////////////////////////////////////////////
// Version History:
//
// Version 2 - 05-Mar-2002
// =======================
// Produced new version with changes as below:
// * added Rich Edit control to available control types
// * added support for ALT+<key> accelerators
// * improved check-box support - now takes lists of strings to use for each
//   of the check states
// * new public functions:
//   - BOOL Create(LPCTSTR lpszText, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID = 0xffff);
//   - virtual BOOL Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext = NULL);
//   - void CheckBoxUseLabel(BOOL bUseLabel = TRUE);
//   - void OffsetCheckbox(BOOL bOffset = TRUE);
//   - void SetCheckText(LPCSTR lpszChecked, LPCSTR lpszUnchecked, LPCSTR lpszIndeterminate = NULL);
//   - void SetCheckText(CStringList& slistChecked, CStringList& slistUnChecked, CStringList& slistIndeterminate);
//   - int GetCheckState(LPCSTR lpszValue);
//   - void SetEditSettings(int nStart, int nEnd, int nFirst);
//   - int GetEditSettings(int& nStart, int &nEnd);
//   - void Initialise();
//   - virtual CWnd* GetMainCtrl();
//   - BOOL IsValueChanged(LPCSTR lpszValue);
//   - CRichEditCtrl* GetRichEditCtrl();
//   - CButton* GetLabelButtonCtrl();
//   - void ModifyAttributes(int nAdd, int nRemove = 0);
//   - int AddListItems(CStringList& list);
//   - int AddListItem(LPCTSTR lpszText);
//
// Version 1 - 27-Jul-2001
// =======================
// Initial version
// 
////////////////////////////////////////////////////////////////////////////
// PLEASE LEAVE THIS HEADER INTACT
////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MultiCtrl.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

enum { ID_LABEL = 1, ID_LABELBTN, ID_STATIC, ID_EDIT, 
	ID_COMBO, ID_BUTTON, ID_DATETIME, ID_SPIN, ID_RICHEDIT };

const DWORD dwStyleCommon = WS_CHILD | WS_TABSTOP | WS_VISIBLE;
const DWORD dwStyleLabel = (dwStyleCommon & ~WS_TABSTOP) | WS_GROUP | SS_NOTIFY;
const DWORD dwStyleLabelButton = dwStyleCommon | BS_PUSHBUTTON;
const DWORD dwStyleStatic = (dwStyleCommon & ~WS_TABSTOP) | SS_NOPREFIX | SS_NOTIFY;
const DWORD dwStyleEdit = dwStyleCommon | ES_AUTOHSCROLL;
const DWORD dwStyleDropDown = dwStyleCommon | WS_BORDER | CBS_AUTOHSCROLL | CBS_DROPDOWNLIST | WS_VSCROLL;
const DWORD dwStyleCombo = dwStyleCommon | WS_BORDER | CBS_AUTOHSCROLL | CBS_DROPDOWN | WS_VSCROLL;
const DWORD dwStyleButton = dwStyleCommon | BS_PUSHBUTTON;
const DWORD dwStyleCheckbox = dwStyleCommon | BS_AUTOCHECKBOX;
const DWORD dwStyleDateTime = dwStyleCommon | WS_BORDER | DTS_SHORTDATEFORMAT | DTS_RIGHTALIGN;
const DWORD dwStyleSpin = dwStyleCommon | WS_BORDER | UDS_AUTOBUDDY | UDS_ALIGNRIGHT;
const DWORD dwStyleRichEdit = dwStyleCommon;

#define DestroyControl(ctrl) \
	if (ctrl != NULL) { ctrl->DestroyWindow(); delete ctrl; ctrl = NULL; }

/*static*/ int CMultiCtrl::ms_nBaseUnitsX = -1;
/*static*/ int CMultiCtrl::ms_nBaseUnitsY = -1;

// macros to convert between pixels and dialog units
#define DLGUNITSX	4	// Dialog units per character horizontally
#define DLGUNITSY	8	// Dialog units per character vertically
#define PixelToDlgUnitX(pixels) (MulDiv((pixels), DLGUNITSX, ms_nBaseUnitsX))
#define PixelToDlgUnitY(pixels) (MulDiv((pixels), DLGUNITSY, ms_nBaseUnitsY))
#define DlgUnitToPixelX(dlus)	(MulDiv((dlus), ms_nBaseUnitsX, DLGUNITSX))
#define DlgUnitToPixelY(dlus)	(MulDiv((dlus), ms_nBaseUnitsY, DLGUNITSY))

// constant for child control sizes in DLUs
#define LABEL_CX	51
#define LABEL_GAP	3 // bit to add to label width when calcing automatically
#define STATIC_CY	8
#define CHECK_CY	10
#define COMBO_CY	92
#define EDIT_CY		13  // same as a standard combo
#define STATICBTN_CX	40
#define BTN_FOCUS_MARGIN  8

/////////////////////////////////////////////////////////////////////////////
// CMultiCtrl

IMPLEMENT_DYNAMIC(CMultiCtrl, CStatic)

CMultiCtrl::CMultiCtrl()
{
  m_pLabel = NULL;
  m_pStatic = NULL;
  m_pLabelBtn = NULL;
  m_pBtn = NULL;
  m_pCombo = NULL;
  m_pEdit = NULL;
  m_pBrowseEdit = NULL;
  m_pDate = NULL;
  m_pSpin = NULL;
  m_pRichEdit = NULL;

  m_nLabelWidth = MCSLW_DEFAULT;

  m_mcType = MCT_NONE;
  m_nAttrib = MCA_NORMAL;

  m_dwStyleAdd = 0;
  m_dwStyleRemove = 0;

  m_bInCreate = FALSE;

  m_nTextLimit = (UINT)-1;

  m_slistChecked.AddTail("1");
  m_slistUnchecked.AddTail("");
  m_slistIndeterminate.AddTail("-1");
  m_bAllowIndeterminate = TRUE;
  m_bOffsetCheckbox = FALSE;
  m_bCheckboxUseLabel = FALSE;
}

CMultiCtrl::~CMultiCtrl()
{
  DestroyControl(m_pLabel);
  DestroyControl(m_pStatic);
  DestroyControl(m_pEdit);
  DestroyControl(m_pBrowseEdit);
  DestroyControl(m_pLabelBtn);
  DestroyControl(m_pCombo);
  DestroyControl(m_pBtn);
  DestroyControl(m_pDate);
  DestroyControl(m_pSpin);
  DestroyControl(m_pRichEdit);
}

BEGIN_MESSAGE_MAP(CMultiCtrl, CStatic)
	//{{AFX_MSG_MAP(CMultiCtrl)
	ON_WM_SIZE()
	ON_WM_SETFOCUS()
	ON_WM_ENABLE()
	ON_WM_KILLFOCUS()
	ON_WM_GETDLGCODE()
	//}}AFX_MSG_MAP
	// make sure we handle notifications from any children
	ON_NOTIFY_EX_RANGE(TTN_NEEDTEXTW, 0, (UINT)-1, OnToolTipNotifyRange)
	ON_NOTIFY_EX_RANGE(TTN_NEEDTEXTA, 0, (UINT)-1, OnToolTipNotifyRange)
END_MESSAGE_MAP()

BOOL CMultiCtrl::Create(LPCTSTR lpszText, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID/*=0xffff*/)
{
  return CStatic::Create(lpszText, dwStyle, rect, pParentWnd, nID);
}

BOOL CMultiCtrl::Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext) 
{
  if (lpszWindowName == NULL)
    lpszWindowName = "";
  if (*lpszWindowName == '\0')
    m_nAttrib |= MCA_NOLABEL;
  m_bInCreate = TRUE;
  BOOL bOK = CWnd::Create(lpszClassName, lpszWindowName, dwStyle, rect, pParentWnd, nID, pContext);
  m_bInCreate = FALSE;
  Initialise();
  return bOK;
}

void CMultiCtrl::CalcBaseUnits(BOOL bRecalc/*=FALSE*/)
{
  // if different instances of this control are likely to have
  // different fonts, then un-comment the next line:
  // bRecalc = TRUE;
  if (bRecalc || ms_nBaseUnitsX == -1 || ms_nBaseUnitsY == -1)
  {
    // get the dialog's dc to do text stuff
    CDC* pDC = GetDC();
    ASSERT(pDC);
    // select the dialog's font into the dc so we can calculate text metrics
    // else we get wrong results as it uses System font by default
    CFont* pOldFont = pDC->SelectObject(GetFont());
    // get relevant system metrics
    // this is what we had in menusupp for doing it:
    // (GetDialogBaseUnits doesn't give correct results as it assumes System font)
    /* apparently this is how Windows does it */
    CSize size = pDC->GetTextExtent("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", 52);
    ms_nBaseUnitsX = size.cx / 52;
    ms_nBaseUnitsY = size.cy;
    pDC->SelectObject(pOldFont);
  }
}

/////////////////////////////////////////////////////////////////////////////
// CMultiCtrl message handlers

void CMultiCtrl::SetLabelText(LPCTSTR lpszString)
{
  // sets the text of the static (label) control
  if (lpszString != NULL)
  {
    CString sText = m_sLabelText = lpszString;
    if (m_mcType == MCT_STATIC)
      sText.Remove('&');  // so no accelerators
    if (m_mcType == MCT_CHECKBOX && m_pBtn != NULL)
      m_pBtn->SetWindowText(sText);
    else if (m_pLabel != NULL)
      m_pLabel->SetWindowText(sText + CString(":"));
    else if (m_pLabelBtn != NULL)
      m_pLabelBtn->SetWindowText(sText + CString("..."));
    if (GetAttributes() & MCA_NOLABEL)
      SetAttributes(GetAttributes() &~ MCA_NOLABEL);
  }
  else
    SetAttributes(GetAttributes() | MCA_NOLABEL);
}

void CMultiCtrl::SetType(MULTICTRLTYPE mct, DWORD dwStyleAdd/*=0*/, DWORD dwStyleRemove/*=0*/)
{
  // switch the controls depending on the type required
  if (mct == m_mcType && dwStyleAdd == m_dwStyleAdd && dwStyleRemove == m_dwStyleRemove)
    return;
  m_mcType = mct;
  m_dwStyleAdd = dwStyleAdd;
  m_dwStyleRemove = dwStyleRemove;
  // if type is static, then remove WS_TABSTOP, else add it
  if (mct == MCT_STATIC)
    ModifyStyle(WS_TABSTOP, 0);
  else if (! (m_nAttrib & MCA_HIDDEN))
    ModifyStyle(0, WS_TABSTOP);
  if (!(GetAttributes() & MCA_NOLABEL) && m_mcType == MCT_STATIC)
    SetLabelText(CString(m_sLabelText));
  CreateControls();
  PositionControls();
  Invalidate();
  UpdateWindow();
}

void CMultiCtrl::OnSize(UINT nType, int cx, int cy)
{
  CStatic::OnSize(nType, cx, cy);
  PositionControls();
}

void CMultiCtrl::PositionControls()
{
  // put children in proper places
  CRect rc;
  GetClientRect(rc);
  int cx = rc.Width();
  int cy = rc.Height();

  // take button width off cx if MCA_BROWSEBTN and type allows browse btn
  int cxButton = 0;
  if (0)  // only on edit control at present
//  if (m_nAttrib & MCA_BROWSEBTN && (m_mcType == MCT_COMBO || 
//	      m_mcType == MCT_DROPLIST || m_mcType == MCT_STATIC || 
//	      m_mcType == MCT_STATICBTN || m_mcType == MCT_DATETIME))
  {
    cxButton = ((m_mcType & MCT_COMBO || m_mcType & MCT_DROPLIST) ? DlgUnitToPixelY(EDIT_CY) : cy);
    cxButton -= 3;
    cx -= cxButton;
  }

  int cxLabel = m_nLabelWidth;
  if (m_nAttrib & MCA_NOLABEL)
    cxLabel = 0;
  else if (m_nLabelWidth == MCSLW_DEFAULT)
    cxLabel = DlgUnitToPixelX(LABEL_CX);
  else if (m_nLabelWidth == MCSLW_AUTO)
    cxLabel = GetLabelWidth(TRUE);

  // label always on left, always same size, middle top and bottom
  if (m_pLabel != NULL)
  {
    m_pLabel->ShowWindow((m_nAttrib & MCA_NOLABEL) ? SW_HIDE : SW_SHOW);
    m_pLabel->MoveWindow(0, (cy - DlgUnitToPixelY(STATIC_CY)) /2, cxLabel, min(cy, DlgUnitToPixelY(STATIC_CY)));
  }

  // combo always on right
  if (m_pCombo != NULL)
    m_pCombo->MoveWindow(cxLabel, 0, cx - cxLabel, DlgUnitToPixelY(COMBO_CY));
  
  // date/time always on right
  if (m_pDate != NULL)	// for some reason the height needs to be incremented by 2
    m_pDate->MoveWindow(cxLabel, 0 - 1, cx - cxLabel, cy + 1);

  // static label button always on left
  if (m_pLabelBtn != NULL)
  {
    if (m_nLabelWidth == MCSLW_AUTO)
      // get text width, and inflate so text is within focus rect
      cxLabel = GetLabelWidth(TRUE) + DlgUnitToPixelX(BTN_FOCUS_MARGIN);
    else if (m_nAttrib & MCA_NOLABEL) // let button fill the control, cos it is the label
      cxLabel = cx;
    else if (m_nLabelWidth == MCSLW_DEFAULT)
      cxLabel = DlgUnitToPixelX(STATICBTN_CX);
    m_pLabelBtn->MoveWindow(0, 0, cxLabel, cy);
    cxLabel += DlgUnitToPixelX(LABEL_GAP);
  }
  
  // static/edit/browseedit on right
  if (m_pStatic != NULL)
  {
    m_pStatic->MoveWindow(cxLabel, (cy - DlgUnitToPixelY(STATIC_CY)) /2, cx - cxLabel, min(cy, DlgUnitToPixelY(STATIC_CY)));
    m_pStatic->ShowWindow(m_pLabelBtn != NULL && m_nAttrib & MCA_NOLABEL ? SW_HIDE : SW_SHOW);
  }
  if (m_pEdit != NULL)
  {
    int cyEdit = min(cy, DlgUnitToPixelY(EDIT_CY));
    if (m_pEdit->GetStyle() & ES_MULTILINE)
      cyEdit = cy;
    m_pEdit->MoveWindow(cxLabel, 0, cx - cxLabel, cyEdit);
  }
  if (m_pRichEdit != NULL)
  {
    int cyEdit = min(cy, DlgUnitToPixelY(EDIT_CY));
    if (m_pRichEdit->GetStyle() & ES_MULTILINE)
      cyEdit = cy;
    m_pRichEdit->MoveWindow(cxLabel, 0, cx - cxLabel, cyEdit);
  }
  if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->MoveWindow(cxLabel, 0, cx - cxLabel, min(cy, DlgUnitToPixelY(EDIT_CY)));

  // btn on left if check box, else on right
  if (m_pBtn != NULL)
  {
    if (m_mcType == MCT_CHECKBOX)
    {
      BOOL bOffset = m_bOffsetCheckbox || m_bCheckboxUseLabel;
      int cxCheck = cx - (bOffset ? cxLabel : 0);
      if (m_bCheckboxUseLabel)
	cxCheck = DlgUnitToPixelY(CHECK_CY);
      m_pBtn->MoveWindow(bOffset ? cxLabel : 0, (cy - DlgUnitToPixelY(CHECK_CY)) / 2, cxCheck, DlgUnitToPixelY(CHECK_CY));
    }
    else
      m_pBtn->MoveWindow(cx, (cy - cxButton) / 2, cxButton, cxButton);
  }

  // reset spin buddy, so the spin adjusts the edit control
  if (m_pSpin != NULL && m_pEdit != NULL)
    m_pSpin->SetBuddy(m_pEdit);
}

MULTICTRLTYPE CMultiCtrl::GetType()
{
  return m_mcType;
}

void CMultiCtrl::Initialise()
{
  CWnd* pParent = GetParent();
  if (pParent != NULL)
    SetFont(pParent->GetFont());

  CalcBaseUnits();

  // set the initial type to a simple edit control
  if (! m_bInCreate)
    SetType(MCT_EDIT);

  if (m_sLabelText.IsEmpty())
    SetAttributes(GetAttributes() | MCA_NOLABEL);
}

void CMultiCtrl::PreSubclassWindow() 
{
  CStatic::PreSubclassWindow();

  // get any text, and use it as the label text
  CStatic::GetWindowText(m_sLabelText);
  // then remove the window text
  CStatic::SetWindowText("");

  // we don't really want any styles here, except WS_CHILD, WS_TABSTOP, WS_GROUP, 
  // and visible if it was already there
  ModifyStyleEx(GetExStyle(), 0);
  ModifyStyle(GetStyle() & ~WS_VISIBLE, WS_CHILD | WS_TABSTOP | WS_GROUP);

  Initialise();
}

BOOL CMultiCtrl::CreateLabel()
{
  if (m_pLabel == NULL)
  {
    m_pLabel = new CStatic;
    if (m_pLabel == NULL || !m_pLabel->Create(m_sLabelText + CString(':'), 
					dwStyleLabel, CRect(0, 0, 0, 0), this, ID_LABEL))
      return FALSE;
  }
  m_pLabel->SetFont(GetFont());
  m_pLabel->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateLabelButton()
{
  if (m_pLabelBtn == NULL)
  {
    m_pLabelBtn = new CButton;
    if (m_pLabelBtn == NULL || !m_pLabelBtn->Create(m_sLabelText + CString("..."), 
				  dwStyleLabelButton, CRect(0, 0, 0, 0), this, ID_LABELBTN))
      return FALSE;
  }
  m_pLabelBtn->SetFont(GetFont());
  m_pLabelBtn->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateStatic()
{
  DWORD dwStyle = (dwStyleStatic &~ m_dwStyleRemove) | m_dwStyleAdd;
  if (m_pStatic != NULL && m_pStatic->GetStyle() != dwStyle)
    DestroyControl(m_pStatic);
  if (m_pStatic == NULL)
  {
    m_pStatic = new CStatic;
    if (m_pStatic == NULL || !m_pStatic->Create("", dwStyle, CRect(0, 0, 0, 0), this, ID_STATIC))
      return FALSE;
  }
  m_pStatic->SetFont(GetFont());
  m_pStatic->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateCombo()
{
  DWORD dwStyleRemove = m_dwStyleRemove | (m_nAttrib & MCA_HISTORY ? CBS_SORT : 0);
  DWORD dwStyleAdd = m_dwStyleAdd & ~(m_nAttrib & MCA_HISTORY ? CBS_SORT : 0);
  DWORD dwStyle = (dwStyleCombo &~ dwStyleRemove) | dwStyleAdd;
  if (m_pCombo != NULL && m_pCombo->GetStyle() != dwStyle)
    DestroyControl(m_pCombo);
  if (m_pCombo == NULL)
  {
    m_pCombo = new CHistoryCombo;
    if (m_pCombo == NULL || !m_pCombo->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_COMBO))
      return FALSE;
  }
  m_pCombo->SetFont(GetFont());
  m_pCombo->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  return TRUE;
}

BOOL CMultiCtrl::CreateDropDown()
{
  DWORD dwStyleRemove = m_dwStyleRemove | (m_nAttrib & MCA_HISTORY ? CBS_SORT : 0);
  DWORD dwStyleAdd = m_dwStyleAdd & ~(m_nAttrib & MCA_HISTORY ? CBS_SORT : 0);
  DWORD dwStyle = (dwStyleDropDown &~ dwStyleRemove) | dwStyleAdd;
  if (m_pCombo != NULL && m_pCombo->GetStyle() != dwStyle)
    DestroyControl(m_pCombo);
  if (m_pCombo == NULL)
  {
    m_pCombo = new CHistoryCombo;
    if (m_pCombo == NULL || !m_pCombo->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_COMBO))
      return FALSE;
  }
  m_pCombo->SetFont(GetFont());
  m_pCombo->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  return TRUE;
}

BOOL CMultiCtrl::CreateEdit()
{
  DWORD dwStyle = (dwStyleEdit &~ m_dwStyleRemove) | 
		      (m_dwStyleAdd | (m_nAttrib & MCA_READONLY ? ES_READONLY : 0));
  if (m_pEdit != NULL && m_pEdit->GetStyle() != dwStyle)
    DestroyControl(m_pEdit);
  if (m_pEdit == NULL)
  {
    m_pEdit = new CEdit;
    if (m_pEdit == NULL || !m_pEdit->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_EDIT))
      return FALSE;
  }
  m_pEdit->ModifyStyleEx(0, WS_EX_CLIENTEDGE | WS_EX_NOPARENTNOTIFY);
  m_pEdit->SetFont(GetFont());
  m_pEdit->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateRichEdit()
{
  DWORD dwStyle = (dwStyleRichEdit &~ m_dwStyleRemove) | 
		      (m_dwStyleAdd | (m_nAttrib & MCA_READONLY ? ES_READONLY : 0));
  if (m_pRichEdit != NULL && m_pRichEdit->GetStyle() != dwStyle)
    DestroyControl(m_pRichEdit);
  if (m_pRichEdit == NULL)
  {
    m_pRichEdit = new CRichEditCtrl;
    if (m_pRichEdit == NULL || !m_pRichEdit->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_RICHEDIT))
      return FALSE;
  }
  m_pRichEdit->ModifyStyleEx(0, WS_EX_CLIENTEDGE | WS_EX_NOPARENTNOTIFY);
  m_pRichEdit->SetFont(GetFont());
  m_pRichEdit->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateBrowseEdit()
{
  DWORD dwStyle = (dwStyleEdit &~ m_dwStyleRemove) | 
		      (m_dwStyleAdd | (m_nAttrib & MCA_READONLY ? ES_READONLY : 0));
  if (m_pBrowseEdit != NULL && m_pBrowseEdit->GetStyle() != dwStyle)
    DestroyControl(m_pBrowseEdit);
  if (m_pBrowseEdit == NULL)
  {
    m_pBrowseEdit = new CBrowseEdit;
    if (m_pBrowseEdit == NULL || !m_pBrowseEdit->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_EDIT))
      return FALSE;
  }
  m_pBrowseEdit->ModifyStyleEx(0, WS_EX_CLIENTEDGE | WS_EX_NOPARENTNOTIFY);
  m_pBrowseEdit->SetFont(GetFont());
  m_pBrowseEdit->EnableWindow(!(m_nAttrib & MCA_DISABLED));
  return TRUE;
}

BOOL CMultiCtrl::CreateCheckbox()
{
  BOOL bResetIndeterminate = TRUE;
  DWORD dwStyle = (dwStyleCheckbox &~ m_dwStyleRemove) | m_dwStyleAdd;
  if (m_pBtn != NULL)
  {
    DWORD dwCurStyle = (m_pBtn->GetStyle() & 0x7L);
    BOOL bCheckbox = dwCurStyle == BS_AUTOCHECKBOX || dwCurStyle == BS_CHECKBOX || 
		      dwCurStyle == BS_AUTO3STATE || dwCurStyle == BS_3STATE;
    if (! bCheckbox)
      m_pBtn->ModifyStyle(0xffffffff, dwStyle);
    else
      bResetIndeterminate = FALSE;
  }
  else
  {
    m_pBtn = new CButton;
    if (m_pBtn == NULL || !m_pBtn->Create("", dwStyle, CRect(0, 0, 0, 0), this, ID_BUTTON))
      return FALSE;
  }
  if (! m_bCheckboxUseLabel)
    m_pBtn->SetWindowText(m_sLabelText);
  m_pBtn->SetFont(GetFont());
  m_pBtn->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  if (bResetIndeterminate)
    m_bAllowIndeterminate = TRUE;
  return TRUE;
}

BOOL CMultiCtrl::CreateButton()
{
  DWORD dwStyle = (dwStyleButton &~ m_dwStyleRemove) | m_dwStyleAdd;
  if (m_pBtn != NULL)
  {
    if (m_pBtn->GetStyle() != dwStyleButton)
      m_pBtn->ModifyStyle(0xffffffff, dwStyle);
  }
  else
  {
    m_pBtn = new CButton;
    if (m_pBtn == NULL || !m_pBtn->Create("", dwStyle, CRect(0, 0, 0, 0), this, ID_BUTTON))
      return FALSE;
  }
  m_pBtn->SetWindowText("...");
  m_pBtn->SetFont(GetFont());
  m_pBtn->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  return TRUE;
}

BOOL CMultiCtrl::CreateDateTime()
{
  DWORD dwStyle = (dwStyleDateTime &~ m_dwStyleRemove) | m_dwStyleAdd;
  if (m_pDate != NULL && m_pDate->GetStyle() != dwStyle)
    DestroyControl(m_pDate);
  if (m_pDate == NULL)
  {
    m_pDate = new CDateTimeCtrl;
    if (m_pDate == NULL || !m_pDate->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_DATETIME))
      return FALSE;
  }
  m_pDate->SetFont(GetFont());
  m_pDate->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  return TRUE;
}

BOOL CMultiCtrl::CreateSpin()
{
  DWORD dwStyle = (dwStyleSpin &~ m_dwStyleRemove) | m_dwStyleAdd;
  if (m_pSpin != NULL && m_pSpin->GetStyle() != dwStyle)
    DestroyControl(m_pSpin);
  if (m_pSpin == NULL)
  {
    m_pSpin = new CSpinButtonCtrl;
    if (m_pSpin == NULL || !m_pSpin->Create(dwStyle, CRect(0, 0, 0, 0), this, ID_SPIN))
      return FALSE;
  }
  m_pSpin->SetFont(GetFont());
  m_pSpin->EnableWindow(!(m_nAttrib & MCA_DISABLED) && !(m_nAttrib & MCA_READONLY));
  return TRUE;
}

BOOL CMultiCtrl::CreateControls()
{
  m_sValueText = GetValueText();

  if (m_nAttrib & MCA_HIDDEN)
  {
    DestroyControl(m_pLabel);
    DestroyControl(m_pStatic);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pEdit);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pCombo);
    DestroyControl(m_pBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    return TRUE;
  }

  // create each of the child controls, depending on the field type
  // the order of creation determines tab order
  switch (m_mcType)
  {
  case MCT_STATIC:
    DestroyControl(m_pEdit);
    DestroyControl(m_pCombo);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabel()) return FALSE;
    if (!CreateStatic()) return FALSE;
    break;

  case MCT_STATICBTN:
    DestroyControl(m_pEdit);
    DestroyControl(m_pCombo);
    DestroyControl(m_pLabel);
    DestroyControl(m_pDate);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabelButton())
      return FALSE;
    if (m_nAttrib & MCA_NOLABEL){
      DestroyControl(m_pStatic); }
    else if (! CreateStatic()) return FALSE;
    break;

  case MCT_EDIT:
  case MCT_UPDOWNEDIT:
    DestroyControl(m_pStatic);
    DestroyControl(m_pCombo);
    DestroyControl(m_pBtn);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pRichEdit);
    if (m_mcType != MCT_UPDOWNEDIT)
      DestroyControl(m_pSpin);
    if (m_mcType == MCT_EDIT && m_nAttrib & MCA_BROWSEBTN)
    {
      DestroyControl(m_pEdit);
      if (!CreateBrowseEdit()) return FALSE;
    }
    else
    {
      DestroyControl(m_pBrowseEdit);
      if (!CreateEdit()) return FALSE;
    }
    if (!CreateLabel()) return FALSE;
    if (m_mcType == MCT_UPDOWNEDIT)
      if (!CreateSpin()) return FALSE;
    break;

  case MCT_RICHEDIT:
    DestroyControl(m_pStatic);
    DestroyControl(m_pEdit);
    DestroyControl(m_pCombo);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabel()) return FALSE;
    if (!CreateRichEdit()) return FALSE;
    break;

  case MCT_COMBO:
    DestroyControl(m_pStatic);
    DestroyControl(m_pEdit);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabel()) return FALSE;
    if (!CreateCombo()) return FALSE;
    break;

  case MCT_DROPLIST:
    DestroyControl(m_pStatic);
    DestroyControl(m_pEdit);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabel()) return FALSE;
    if (!CreateDropDown()) return FALSE;
    break;

  case MCT_CHECKBOX:
    DestroyControl(m_pStatic);
    DestroyControl(m_pCombo);
    if (! m_bCheckboxUseLabel)
      DestroyControl(m_pLabel);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pEdit);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (!CreateCheckbox()) return FALSE;
    break;

  case MCT_DATETIME:
    DestroyControl(m_pStatic);
    DestroyControl(m_pCombo);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pEdit);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
    if (m_nAttrib & MCA_BROWSEBTN){ 
      if (!CreateButton()) return FALSE; }
    else
      DestroyControl(m_pBtn);
    if (!CreateLabel()) return FALSE;
    if (!CreateDateTime()) return FALSE;
    break;

  default:
    DestroyControl(m_pLabel);
    DestroyControl(m_pStatic);
    DestroyControl(m_pEdit);
    DestroyControl(m_pBrowseEdit);
    DestroyControl(m_pLabelBtn);
    DestroyControl(m_pCombo);
    DestroyControl(m_pBtn);
    DestroyControl(m_pDate);
    DestroyControl(m_pSpin);
    DestroyControl(m_pRichEdit);
  }

  // set the value text to what it used to be
  SetValueText(m_sValueText);

  SetTextLimit(m_nTextLimit);

  PositionControls();

  SetHelpString(m_sHelpString);

  return TRUE;
}

void CMultiCtrl::OnSetFocus(CWnd* pOldWnd) 
{
  // set the focus to the most appropriate child depending on the type
  if (m_pEdit != NULL)
  {
    m_pEdit->SetFocus();
    m_pEdit->SetSel(0, -1);
  }
  if (m_pRichEdit != NULL)
  {
    m_pRichEdit->SetFocus();
    m_pRichEdit->SetSel(0, -1);
  }
  if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->SetFocus();
  else if (m_pCombo != NULL)
    m_pCombo->SetFocus();
  else if (m_pDate != NULL)
    m_pDate->SetFocus();
  else if (m_mcType == MCT_CHECKBOX && m_pBtn != NULL)
    m_pBtn->SetFocus();
  else if (m_pLabelBtn != NULL)
    m_pLabelBtn->SetFocus();
  CStatic::OnSetFocus(pOldWnd);

  // set the main frame's message text to the help string
  SetStatusText(m_sHelpString);
}

void CMultiCtrl::OnEnable(BOOL bEnable) 
{
  CStatic::OnEnable(bEnable);
  
  // set the attrib to the appropriate type
  int nAttrib = (int)GetAttributes();
  if (bEnable)
    nAttrib &= ~MCA_DISABLED;
  else
    nAttrib |= MCA_DISABLED;
  SetAttributes((MULTICTRLATTRIB)nAttrib);
}

void CMultiCtrl::SetFont(CFont *pFont, BOOL bRedraw)
{
  CStatic::SetFont(pFont, bRedraw);
  // set the font of each child control
  if (m_pLabel != NULL)
    m_pLabel->SetFont(pFont, bRedraw);
  if (m_pStatic != NULL)
    m_pStatic->SetFont(pFont, bRedraw);
  if (m_pLabelBtn != NULL)
    m_pLabelBtn->SetFont(pFont, bRedraw);
  if (m_pEdit != NULL)
    m_pEdit->SetFont(pFont, bRedraw);
  if (m_pRichEdit != NULL)
    m_pRichEdit->SetFont(pFont, bRedraw);
  if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->SetFont(pFont, bRedraw);
  if (m_pCombo != NULL)
    m_pCombo->SetFont(pFont, bRedraw);
  if (m_pBtn != NULL)
    m_pBtn->SetFont(pFont, bRedraw);
  if (m_pDate != NULL)
    m_pDate->SetFont(pFont, bRedraw);
}

void CMultiCtrl::SetWindowText(LPCTSTR lpszText)
{
  SetValueText(lpszText);
}

void CMultiCtrl::GetWindowText(CString &rText) const
{
  rText = GetValueText();
}

int CMultiCtrl::GetWindowText(LPTSTR lpszBuffer, int nMax) const
{
  CString sText = GetValueText();
  strncpy(lpszBuffer, sText, nMax);
  return sText.GetLength();
}

int CMultiCtrl::GetWindowTextLength() const
{
  CString sText = GetValueText();
  return sText.GetLength();
}

BOOL IsInStringList(CStringList& list, LPCSTR lpsz, BOOL bMatchCase = FALSE)
{
  if (lpsz == NULL)
    return FALSE;
  POSITION pos = list.GetHeadPosition();
  while (pos != NULL)
  {
    CString sItem = list.GetNext(pos);
    if (bMatchCase)
    {
      if (sItem == lpsz)
	return TRUE;
    }
    else
    {
      if (sItem.CompareNoCase(lpsz) == 0)
	return TRUE;
    }
  };
  return FALSE;
}

// return the check box check state corresponding to the value text
int CMultiCtrl::GetCheckState(LPCSTR lpszValue)
{
  int nCheck = BST_UNCHECKED;
  if (IsInStringList(m_slistUnchecked, lpszValue))
    nCheck = BST_UNCHECKED;
  else if (IsInStringList(m_slistChecked, lpszValue))
    nCheck = BST_CHECKED;
  else if (IsInStringList(m_slistIndeterminate, lpszValue))
    nCheck = BST_INDETERMINATE;
  return nCheck;
}

void CMultiCtrl::SetValueText(LPCTSTR lpszValue)
{
  m_sValueText = lpszValue; // keep it so it's used when controls are created
  if (lpszValue == NULL)
    return;

  if (m_pCombo != NULL)
  {
    if (m_sValueText.IsEmpty())	// simply unselect the selection
      m_pCombo->SetCurSel(-1);
    else
    {
      // find it in the list
      int nIndex = m_pCombo->FindStringExact(-1, lpszValue);
      // try and select it, else add if combo
      m_pCombo->SetCurSel(nIndex);
      if (nIndex == -1 && m_mcType == MCT_COMBO)
	m_pCombo->SetWindowText(lpszValue);
    }
  }
  else if (m_pStatic != NULL)
    m_pStatic->SetWindowText(lpszValue);
  else if (m_pEdit != NULL)
    m_pEdit->SetWindowText(lpszValue);
  else if (m_pRichEdit != NULL)
    m_pRichEdit->SetWindowText(lpszValue);
  else if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->SetWindowText(lpszValue);
  else if (m_pBtn != NULL && m_mcType == MCT_CHECKBOX)
  {
    BOOL b3State = ((m_pBtn->GetStyle() & BS_AUTO3STATE) == BS_AUTO3STATE) || 
		    ((m_pBtn->GetStyle() & BS_3STATE) == BS_3STATE);
    int nCheck = GetCheckState(lpszValue);
    if (nCheck == BST_INDETERMINATE)
    {
      if (! b3State)
      {
	m_pBtn->ModifyStyle(0x7L, BS_AUTO3STATE); //must make sure lower 3 bits are cleared
	TRACE("WARNING: Setting non-tristate checkbox to indeterminate!\n");
	m_bAllowIndeterminate = FALSE;
      }
    }
    m_pBtn->SetCheck(nCheck);
  }
  else if (m_pDate != NULL)
  {
    // make a time object
    if (*lpszValue != '\0')
    {
      COleDateTime dt;
      dt.ParseDateTime(lpszValue);
      if (dt.GetStatus() == COleDateTime::valid)
	m_pDate->SetTime(dt);
      else
	m_pDate->SetTime();
#if 0
      // TODO: this assumes a date string format of yyyy/mm/dd
      SYSTEMTIME st;
      ZeroMemory(&st, sizeof(SYSTEMTIME));
      LPSTR lpszVal = strdup(lpszValue);
      LPSTR tok = strtok(lpszVal, "/");
      if (tok != NULL && ! streql(tok, lpszVal))
      {
	st.wYear = atoi(tok);
	tok = strtok(NULL, "/");
	st.wMonth = atoi(tok);
	tok = strtok(NULL, "/");
	st.wDay = atoi(tok);
	m_pDate->SetTime(&st);
      }
      free(lpszVal);
#endif
    }
    else // set no time (turn off checkbox)
      m_pDate->SetTime();
  }
}

CString CMultiCtrl::GetValueText() const
{
  CString sText("");
  if (m_pCombo != NULL)
  {
    // get current selection
    int nIndex = m_pCombo->GetCurSel();
    // if it's ok, use it
    if (nIndex == -1)
      m_pCombo->GetWindowText(sText);
    else
      m_pCombo->GetLBText(nIndex, sText);
  }
  else if (m_pStatic != NULL)
    m_pStatic->GetWindowText(sText);
  else if (m_pEdit != NULL)
    m_pEdit->GetWindowText(sText);
  else if (m_pRichEdit != NULL)
    m_pRichEdit->GetWindowText(sText);
  else if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->GetWindowText(sText);
  else if (m_pBtn != NULL && m_mcType == MCT_CHECKBOX)
  {
    int nCheck = m_pBtn->GetCheck();
    sText = (nCheck == BST_INDETERMINATE) ? m_slistIndeterminate.GetHead() : 
	    (nCheck == BST_CHECKED) ? m_slistChecked.GetHead() : m_slistUnchecked.GetHead();
  }
  else if (m_pDate != NULL)
  {
    m_pDate->GetWindowText(sText);
    COleDateTime dt;
    m_pDate->GetTime(dt);
    sText = dt.Format();
#if 0
    CTime time;
    m_pDate->GetTime(time);
    // TODO: this returns date in format yyyy/mm/dd
    sText = time.FormatGmt("%Y/%m/%d");
#endif
  }
  else
    return m_sValueText;
  return sText;
}

// add item to list, if the control has a list
// and return the index of the item, or -1 if error
int CMultiCtrl::AddListItem(LPCTSTR lpszText)
{
  if (lpszText == NULL)
    return -1;
  if (m_mcType != MCT_COMBO && m_mcType != MCT_DROPLIST)
    return -1;
  return m_pCombo->AddString(lpszText);
}

// add multiple items to list, if the control has a list
// and return number of items added, or -1 if error
int CMultiCtrl::AddListItems(CStringList &list)
{
  if (m_mcType != MCT_COMBO && m_mcType != MCT_DROPLIST)
    return -1;
  POSITION pos = list.GetHeadPosition();
  while (pos != NULL)
  {
    CString sText = list.GetNext(pos);
    m_pCombo->AddString(sText);
  }
  return list.GetCount();
}

void CMultiCtrl::ModifyAttributes(int nAdd, int nRemove/*=0*/)
{
  int nAttrib = GetAttributes();
  nAttrib &= ~nRemove;
  nAttrib |= nAdd;
  SetAttributes(nAttrib);
}

void CMultiCtrl::SetAttributes(int nAttrib)
{
  if (m_nAttrib == nAttrib)
    return;
  m_nAttrib = nAttrib;
  // if it's hidden or disabled, remove tabstop
  if (m_nAttrib & MCA_HIDDEN || m_nAttrib & MCA_DISABLED)
    ModifyStyle(WS_TABSTOP, 0);
  else if (m_mcType != MCT_STATIC)
    ModifyStyle(0, WS_TABSTOP);
  CreateControls();
}

int CMultiCtrl::GetAttributes()
{
  return m_nAttrib;
}

BOOL CMultiCtrl::OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult) 
{
  NMHDR* pnmhdr = (NMHDR*)lParam;
  if (pnmhdr == NULL || (pnmhdr->code != TTN_NEEDTEXTA && pnmhdr->code != TTN_NEEDTEXTW))
  {
    // pass the notifications from the controls to the parent
    CWnd* pParent = GetParent();
    if (pParent != NULL)
    {
      pnmhdr->idFrom = GetDlgCtrlID();
      pnmhdr->hwndFrom = GetSafeHwnd();
      *pResult = pParent->SendMessage(WM_NOTIFY, (WPARAM)GetDlgCtrlID(), lParam);
      return TRUE;
    }
  }
  return CStatic::OnNotify(wParam, lParam, pResult);
}

BOOL CMultiCtrl::OnCommand(WPARAM wParam, LPARAM lParam) 
{
  // if it's a notification from the child that if has got/lost the focus, then
  // set/clear the message text
  // works OK for edits, buttons, and combos
  // TODO: do something that'll work for dates/spins
  UINT nNotifyCode = HIWORD(wParam);
  if (nNotifyCode == EN_SETFOCUS || 
      nNotifyCode == CBN_SETFOCUS || 
      nNotifyCode == BN_SETFOCUS)
    // set the main frame's message text to the help string
    SetStatusText(m_sHelpString);
  else if (nNotifyCode == EN_KILLFOCUS || 
	    nNotifyCode == CBN_KILLFOCUS || 
	    nNotifyCode == BN_KILLFOCUS)
    // set the main frame's message text to the help string
    SetStatusText();

  // if clicking on a non-tristate checkbox, then make sure we can't
  // go to indeterminate state
  if (m_pBtn != NULL && LOWORD(wParam) == ID_BUTTON && (nNotifyCode == BN_CLICKED || nNotifyCode == 0) && 
//	(! ((m_dwStyleAdd & BS_AUTOCHECKBOX) == BS_AUTO3STATE || (m_dwStyleAdd & BS_3STATE) == BS_3STATE)) && 
	  ((m_pBtn->GetStyle() & BS_AUTO3STATE) == BS_AUTO3STATE || (m_pBtn->GetStyle() & BS_3STATE) == BS_3STATE))
  {
    // don't allow indeterminate state
    if (! m_bAllowIndeterminate && m_pBtn->GetCheck() == BST_INDETERMINATE)
    {
      // send button click to button again, to push on to next state
      //m_pBtn->SendMessage(WM_COMMAND, wParam, lParam);
      m_pBtn->SetCheck(BST_UNCHECKED);
    }
  }

  // pass the commands from the controls to the parent
  CWnd* pParent = GetParent();
  if (pParent != NULL)
  {
    // we don't want to notify parent if static clicked, when control is showing label btn
    if (LOWORD(wParam) == ID_STATIC || LOWORD(wParam) == ID_LABEL)
      return TRUE;
    if (!pParent->SendMessage(WM_COMMAND, MAKEWPARAM(GetDlgCtrlID(), HIWORD(wParam)), (LPARAM)GetSafeHwnd()))
      return TRUE;
  }
  return CStatic::OnCommand(wParam, lParam);
}

void CMultiCtrl::SetLabelWidth(int nWidth/*=MCSLW_DEFAULT*/)
{
  m_nLabelWidth = nWidth;
  PositionControls();
}

int CMultiCtrl::GetLabelWidth(BOOL bAuto/*=FALSE*/)
{
  if (bAuto || m_nLabelWidth == MCSLW_AUTO)
  {
    // get the size of the label, based on its contents
    // then add a small gap
    if (!m_sLabelText.IsEmpty())
    {
      CDC* pDC = GetDC();
      CFont* pFont = GetFont();
      if (pDC != NULL && pFont != NULL)
      {
	int cxLabel;
	CFont* pOldFont = pDC->SelectObject(pFont);
	CRect rc(0, 0, 0, 0);
	pDC->DrawText(m_sLabelText + CString(':'), rc, DT_CALCRECT | DT_SINGLELINE);
	pDC->SelectObject(pOldFont);
	cxLabel = rc.Width() + DlgUnitToPixelX(LABEL_GAP);
	return cxLabel;
      }
      else
      {
	if (m_mcType == MCT_STATICBTN && m_nAttrib & MCA_NOLABEL)
	{
	  // label/button fills control in this instance
	  CRect rc;
	  GetWindowRect(&rc);
	  m_nLabelWidth = rc.Width();
	}
	else
	  m_nLabelWidth = (m_mcType == MCT_STATICBTN ? DlgUnitToPixelX(STATICBTN_CX) : DlgUnitToPixelX(LABEL_CX));
      }
    }
    else
      m_nLabelWidth = 0;
  }
  if (m_nLabelWidth == MCSLW_DEFAULT)
  {
    if (m_mcType == MCT_STATICBTN && m_nAttrib & MCA_NOLABEL)
    {
      // label/button fills control in this instance
      CRect rc;
      GetWindowRect(&rc);
      return rc.Width();
    }
    return (m_mcType == MCT_STATICBTN ? DlgUnitToPixelX(STATICBTN_CX) : DlgUnitToPixelX(LABEL_CX));
  }
  return m_nLabelWidth;
}

#define BEGIN_CTRL_TYPE_MAP(type)	CString sTypeName; switch(type) {
#define END_CTRL_TYPE_MAP		} return sTypeName;
#define MAP_CTRL_TYPE(type)		case type: sTypeName = #type; break

CString CMultiCtrl::GetTypeName()
{
  BEGIN_CTRL_TYPE_MAP(m_mcType)
    MAP_CTRL_TYPE(MCT_NONE);
    MAP_CTRL_TYPE(MCT_STATIC);
    MAP_CTRL_TYPE(MCT_STATICBTN);
    MAP_CTRL_TYPE(MCT_EDIT);
    MAP_CTRL_TYPE(MCT_UPDOWNEDIT);
    MAP_CTRL_TYPE(MCT_DROPLIST);
    MAP_CTRL_TYPE(MCT_COMBO);
    MAP_CTRL_TYPE(MCT_CHECKBOX);
    MAP_CTRL_TYPE(MCT_DATETIME);
    MAP_CTRL_TYPE(MCT_RICHEDIT);
  END_CTRL_TYPE_MAP
}

void CMultiCtrl::ResetContent()
{
  // clear the value text, and clear the list contents for combos
  SetValueText("");
  if (m_pCombo != NULL)
    m_pCombo->ResetContent();
}

void CMultiCtrl::EnableControl(BOOL bEnable/*=TRUE*/)
{
  // set the attrib to the appropriate type
  int nAttrib = (int)GetAttributes();
  if (bEnable)
    nAttrib &= ~MCA_DISABLED;
  else
    nAttrib |= MCA_DISABLED;
  SetAttributes((MULTICTRLATTRIB)nAttrib);
}

void CMultiCtrl::SetTextLimit(UINT nMax)
{
  // set the max length of the field's edit control
  if (m_pEdit != NULL)
    m_pEdit->SetLimitText(nMax);
  else if (m_pRichEdit != NULL)
    m_pRichEdit->LimitText(nMax);
  else if (m_pCombo != NULL && m_mcType == MCT_COMBO)
    m_pCombo->LimitText(nMax);
  m_nTextLimit = nMax;
}

// return TRUE if value changed by user since 
// last call to SetValueText, else FALSE
// if a value is specified, simply 
// return whether we currently have that value
BOOL CMultiCtrl::IsValueChanged(LPCSTR lpszValue/*=NULL*/)
{
  if (lpszValue != NULL)
  {
    if (m_mcType == MCT_CHECKBOX)
      return (GetCheckState(lpszValue) != m_pBtn->GetCheck());
    return (GetValueText() != CString(lpszValue));
  }
  return (GetValueText() != m_sValueText);
}

void CMultiCtrl::SetHelpString(LPCSTR lpszHelpString)
{
  BOOL bEnable = (lpszHelpString != NULL && *lpszHelpString != '\0');
  if (bEnable)
    m_sHelpString = lpszHelpString;
  else
    m_sHelpString = "";

  // turn on/off tool-tips for children
  EnableToolTips(bEnable);

  // turn on/off tooltips for each of the child controls as well
  if (m_pBrowseEdit != NULL)
    m_pBrowseEdit->EnableToolTips(bEnable);
#if 0
  if (m_pBtn != NULL)
    m_pBtn->EnableToolTips(bEnable);
  if (m_pCombo != NULL)
    m_pCombo->EnableToolTips(bEnable);
  if (m_pDate != NULL)
    m_pDate->EnableToolTips(bEnable);
  if (m_pEdit != NULL)
    m_pEdit->EnableToolTips(bEnable);
  if (m_pRichEdit != NULL)
    m_pRichEdit->EnableToolTips(bEnable);
  if (m_pLabel != NULL)
    m_pLabel->EnableToolTips(bEnable);
  if (m_pLabelBtn != NULL)
    m_pLabelBtn->EnableToolTips(bEnable);
  if (m_pSpin != NULL)
    m_pSpin->EnableToolTips(bEnable);
  if (m_pStatic != NULL)
    m_pStatic->EnableToolTips(bEnable);
#endif
}

BOOL CMultiCtrl::OnToolTipNotifyRange(UINT nID, NMHDR *pNMHDR, LRESULT *pResult)
{
  // need to handle both ANSI and UNICODE versions of the message
  TOOLTIPTEXTA* pTTTA = (TOOLTIPTEXTA*)pNMHDR;
  TOOLTIPTEXTW* pTTTW = (TOOLTIPTEXTW*)pNMHDR;

  // no point in showing tool-tip for hidden or 'none' controls
  if (m_mcType == MCT_NONE || m_nAttrib & MCA_HIDDEN)
    return FALSE;

  if (pNMHDR->code == TTN_NEEDTEXTA)
    lstrcpyn(pTTTA->szText, m_sHelpString, sizeof(pTTTA->szText));
  else
    _mbstowcsz(pTTTW->szText, m_sHelpString, sizeof(pTTTW->szText));
  *pResult = 0;

  return FALSE;	// we handled it, but allow parent a shot at it
}

// returns the previously set tool-tip/status bar text
CString CMultiCtrl::GetHelpString()
{
  return m_sHelpString;
}

void CMultiCtrl::OnKillFocus(CWnd* pNewWnd) 
{
  CStatic::OnKillFocus(pNewWnd);
  SetStatusText();
}

void CMultiCtrl::SetStatusText(LPCSTR lpszText/*=NULL*/)
{
  // set the main frame's message text to the help string
  CFrameWnd* pFrame = DYNAMIC_DOWNCAST(CFrameWnd, AfxGetMainWnd());
  if (pFrame != NULL)
  {
    if (lpszText == NULL || *lpszText == '\0')
      pFrame->SetMessageText(AFX_IDS_IDLEMESSAGE);
    else
      pFrame->SetMessageText(lpszText);
  }
}

CWnd* CMultiCtrl::GetMainCtrl()
{
  return m_pCombo != NULL ? m_pCombo : 
	  m_pEdit != NULL ? m_pEdit : 
	  m_pBrowseEdit != NULL ? m_pBrowseEdit : 
	  m_pRichEdit != NULL ? m_pRichEdit : 
	  m_pBtn != NULL ? m_pBtn : 
	  m_pStatic != NULL ? m_pStatic : 
	  m_pDate != NULL ? (CWnd*)m_pDate : (CWnd*)NULL;
}

int CMultiCtrl::GetEditSettings(int &nStart, int &nEnd)
{
  int nFirst = 0;
  CEdit* pEdit = GetEditCtrl();
  if (pEdit != NULL)
  {
    pEdit->GetSel(nStart, nEnd);
    nFirst = pEdit->GetFirstVisibleLine();
  }
  else
  {
    CRichEditCtrl* pEdit = GetRichEditCtrl();
    if (pEdit != NULL)
    {
      long nStartChar, nEndChar;
      pEdit->GetSel(nStartChar, nEndChar);
      nStart = nStartChar;
      nEnd = nEndChar;
      nFirst = pEdit->GetFirstVisibleLine();
    }
  }
  return nFirst;
}

void CMultiCtrl::SetEditSettings(int nStart, int nEnd, int nFirst)
{
  CEdit* pEdit = GetEditCtrl();
  if (pEdit != NULL)
  {
    pEdit->SetSel(nStart, nEnd, TRUE);
    pEdit->LineScroll(nFirst);
  }
  else
  {
    CRichEditCtrl* pEdit = GetRichEditCtrl();
    if (pEdit != NULL)
    {
      pEdit->SetSel(nStart, nEnd);
      pEdit->LineScroll(nFirst);
    }
  }
}

inline void CMultiCtrl::SetCheckText(CStringList& slistChecked, CStringList& slistUnchecked, CStringList& slistIndeterminate)
{
  m_slistChecked.RemoveAll();
  m_slistChecked.AddTail(&slistChecked);
  m_slistUnchecked.RemoveAll();
  m_slistUnchecked.AddTail(&slistUnchecked);
  m_slistIndeterminate.RemoveAll();
  m_slistIndeterminate.AddTail(&slistIndeterminate);
}

inline void CMultiCtrl::SetCheckText(LPCSTR lpszChecked, LPCSTR lpszUnchecked, LPCSTR lpszIndeterminate/*=NULL*/)
{
  m_slistChecked.RemoveAll();
  m_slistChecked.AddTail(lpszChecked);
  m_slistUnchecked.RemoveAll();
  m_slistUnchecked.AddTail(lpszUnchecked);
  m_slistIndeterminate.RemoveAll();
  m_slistIndeterminate.AddTail(lpszIndeterminate);
}

inline void CMultiCtrl::OffsetCheckbox(BOOL bOffset/*=TRUE*/)
{
  m_bOffsetCheckbox = bOffset;
}

inline void CMultiCtrl::CheckboxUseLabel(BOOL bUseLabel/*=TRUE*/)
{
  m_bCheckboxUseLabel = bUseLabel;
}

UINT CMultiCtrl::OnGetDlgCode() 
{
  // we want to respond to accelerator keys for our labels.
  // By default, a static control will return DLGC_STATIC
  // which means it doesn't really get anything, so we
  // return a more appropriate code, depending on the type 
  // of the control, so windows will set the focus to it, 
  // or do other appropriate actions.
  // For most controls we respond with DLGC_HASSETSEL, so Windows
  // will set the focus to the control, but for buttons (including)
  // checkboxes) we respond with DLGC_BUTTON | DLGC_UNDEFPUSHBUTTON, 
  // so Windows will 'press' the button (we actually get the code 
  // from the button controls, so we know it's going to be correct)
  int nCode = DLGC_HASSETSEL;
  if (m_mcType == MCT_STATICBTN && m_pLabelBtn != NULL)
    nCode = m_pLabelBtn->SendMessage(WM_GETDLGCODE);
  else if (m_mcType == MCT_CHECKBOX && m_pBtn != NULL)
    nCode = m_pBtn->SendMessage(WM_GETDLGCODE);
  return nCode;
}

LRESULT CMultiCtrl::WindowProc(UINT message, WPARAM wParam, LPARAM lParam) 
{
  if (message == WM_GETTEXT)
  {
    // copy in the label text - this should make accelerators work
    // wParam is sizeof buffer, lParam is buffer (char*)
    LPSTR lpszBuf = (LPSTR)lParam;
    int nBufLen = wParam;
    int nRet = 0;
    if (lpszBuf != NULL && nBufLen > 0)
    {
      CString sText = m_sLabelText;
      if (m_mcType == MCT_STATIC || m_mcType == MCT_NONE)
	sText.Remove('&');  // so no accelerators
      strncpy(lpszBuf, sText, nBufLen);
      lpszBuf[nBufLen - 1] = '\0';
      nRet = strlen(lpszBuf);
    }
    return nRet;
  }
  else if (message == EM_SETSEL)
  {
    // pass this message on to the edit control if there is one
    if (m_pEdit != NULL)
      m_pEdit->SendMessage(EM_SETSEL, wParam, lParam);
    else if (m_pRichEdit != NULL)
      m_pRichEdit->SendMessage(EM_SETSEL, wParam, lParam);
    return 0;
  }
  return CStatic::WindowProc(message, wParam, lParam);
}




